/*
This file can be used to make a JNI library, used by N-View for sending and receiving ICMP packets
on UNIX like operating systems.
Requirements:
- a C/C++ compiler,
- a proper installed J2SDK, version 1.5.0 or above,
- and the files "JPingImp_Unix.c", "de_NView_icmp_JPing.h", provided in this package.

For the following example it is assumed, the j2sdk is installed in the directory "/opt/jdk1.5.0_01/"
and a "gcc" compiler is used. The library for example used by Solaris on a Sparc architecture
can be made by the following command:

gcc -G -I /opt/jdk1.5.0_01/include/ -I /opt/jdk1.5.0_01/include/solaris -o libJPingImp_SunOS_sparc.so JPingImp_Unix.c

The name of the library to be made, depends on your operating system and the architecture as follows:

<library prefix>JPingImp_<operating system>_<architecture>.<library postfix>

You can see the filename, requested by N-View, on startup in the first line of its console output.
*/

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>


#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <sys/fcntl.h>
#include <fcntl.h>

#if !defined (INADDR_NONE)
#define INADDR_NONE (-1)
#endif

#include "de_NView_icmp_JPing.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ICMP_ECHOREPLY                 0
#define ICMP_UNREACHABLE               3
#define ICMP_ECHO                      8
#define ICMP_CODE_NETWORK_UNREACHABLE  0
#define ICMP_CODE_HOST_UNREACHABLE     1

#define STATUS_FAILED       0xFFFF
#define PACKET_SIZE         32
#define ICMP_MIN            8

#if !defined(__GLIBC__) || (__GLIBC__ < 2)
#ifdef __linux__
#include "pingpack.h"
#endif /* __linux__ */
#endif /* __GLIBC__ */

#define SET_NONBLOCKING(fd){int flags=fcntl(fd,F_GETFL);flags|=O_NONBLOCK;fcntl(fd,F_SETFL,flags);}

long GetTickCount(){
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return(tv.tv_sec*1000+(tv.tv_usec/1000));
}

typedef struct
{
  unsigned int   h_len:4;
  unsigned int   version:4;
  unsigned char  tos;
  unsigned short total_len;
  unsigned short ident;
  unsigned short frag_and_flags;
  unsigned char  ttl;
  unsigned char  proto;
  unsigned short checksum;
  unsigned int   sourceIP;
  unsigned int   destIP;
}IP_HEADER;

typedef struct
{
  char           i_type;
  char           i_code;
  unsigned short i_cksum;
  unsigned short i_id;
  unsigned short i_seq;
  unsigned long  timestamp;
}ICMP_HEADER;

//Return:   delay (ms) on success,
//          -1L  UNREACHABLE
//          -2L  invalid/foreign Packet
long DecodePacket(char *buf,int bytes,unsigned short myId){
    IP_HEADER *ipHeader;
    ICMP_HEADER *icmpHeader;
    unsigned short ipHeaderLen;
    long rc;

    ipHeader=(IP_HEADER*)buf;
//    ipHeaderLen=ipHeader->h_len*4;
    ipHeaderLen=20;
    if(bytes<ipHeaderLen+ICMP_MIN){
        printf("Too few bytes\n");
        return -2L;
    }
    icmpHeader=(ICMP_HEADER*)(buf+ipHeaderLen);
    if(icmpHeader->i_id!=(getpid()&0x0FFFF)){   //not my packet id
        return -2L;
    }
    if(icmpHeader->i_seq!=myId){                //not my packet seqnum
        return -2L;
    }
    if(icmpHeader->i_type!=ICMP_ECHOREPLY){
        if(icmpHeader->i_type==ICMP_UNREACHABLE){
            if(icmpHeader->i_code==ICMP_CODE_HOST_UNREACHABLE){
                return -1L;
            }
            if(icmpHeader->i_code==ICMP_CODE_NETWORK_UNREACHABLE){
                return -1L;
            }
        }
        else{return -2L;}   //invalid type
    }
    rc=GetTickCount()-icmpHeader->timestamp;
    return rc;
}

unsigned short checksum(unsigned short *buffer,int size){
    unsigned long cksum=0;
    while(size>1){
        cksum+=*buffer++;
        size-=sizeof(unsigned short);
    }
    if(size){cksum+=*(unsigned char*)buffer;}
    cksum=(cksum>>16)+(cksum&0xffff);
    cksum+=(cksum>>16);
    return (unsigned short)(~cksum);
}

void SetupPacket(char * icmpdata,int datasize,unsigned short myId){
    ICMP_HEADER *icmp_hdr;
    char *datapart;

    icmp_hdr=(ICMP_HEADER*)icmpdata;
    icmp_hdr->i_type=ICMP_ECHO;
    icmp_hdr->i_code=0;
    icmp_hdr->i_id=getpid()&0xFFFF;
    icmp_hdr->i_cksum=0;
    icmp_hdr->i_seq=myId;
    datapart=icmpdata+sizeof(ICMP_HEADER);
    memset(datapart,'!',datasize-sizeof(ICMP_HEADER));
    memcpy(&datapart[3],"N-View",6);
}

long sendPacket(const char *str,unsigned short Id,int sockRaw){
    struct sockaddr_in addrDest;
    struct hostent Hostent;
    struct hostent* ptrHostent = &Hostent;
	char hbuf[4096];
	int herr;
    char icmpdata[PACKET_SIZE+sizeof(ICMP_HEADER)];
    unsigned int addr=0;

#ifdef __linux__
	if (gethostbyname_r(str, ptrHostent, hbuf, sizeof(hbuf), &ptrHostent, &herr) == NULL) {
#else /* __linux__ */
	if (gethostbyname_r(str, ptrHostent, hbuf, sizeof(hbuf), &herr) == NULL) {
#endif /* __linux__ */
		if ((addr = inet_addr(str)) == -1) {
			printf("Error: Cannot resolve Host: %s\n",str);
			close(sockRaw);
			return -1L;
		}
        addrDest.sin_addr.s_addr=addr;
        addrDest.sin_family=AF_INET;
    } else {
        memcpy((void *)&(addrDest.sin_addr), (void *)ptrHostent->h_addr,ptrHostent->h_length);
        addrDest.sin_family=ptrHostent->h_addrtype;
	}

    SetupPacket(icmpdata,PACKET_SIZE+sizeof(ICMP_HEADER),Id);
    ((ICMP_HEADER*)icmpdata)->i_cksum=0;
    ((ICMP_HEADER*)icmpdata)->timestamp=GetTickCount();
    ((ICMP_HEADER*)icmpdata)->i_cksum=checksum((unsigned short*)icmpdata,PACKET_SIZE+sizeof(ICMP_HEADER));
//printf("SENDING: %d == %d\n",((ICMP_HEADER*)icmpdata)->i_seq,Id);
    return sendto(sockRaw,icmpdata,PACKET_SIZE+sizeof(ICMP_HEADER),0,(struct sockaddr *)&addrDest,sizeof(addrDest));
}

JNIEXPORT jlong JNICALL 
Java_de_NView_icmp_JPing_sockStartup(JNIEnv *env,jobject obj){
    int sockRaw;
    int bytesSent;

    sockRaw=socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);
    if(sockRaw<0){
        printf("Error: Cannot create Socket\n");
        return -1L;
    }
    bytesSent=sendPacket("172.0.0.1",0,sockRaw);
    if(bytesSent<0){
        printf("Error: Cannot send packet\n");
        close(sockRaw);
        return -1L;
    }
    return 0L;
}

JNIEXPORT jlong JNICALL 
Java_de_NView_icmp_JPing_jPing(JNIEnv *env,jobject obj,jint timeoutMillis,jstring addrString,jint Id){
    const char* str;
    int sockRaw;
    struct sockaddr_in addrFrom;
    int addrFromLen=sizeof(addrFrom);
    char recvbuf[PACKET_SIZE+sizeof(IP_HEADER)+sizeof(ICMP_HEADER)];
    int bytesReceived,bytesSent,remainingTime;
    unsigned long timeout;
    struct timeval tv;
    fd_set rd,wr;
    long rc;

    if(addrString==NULL){
        printf("Error: NULL addrString!\n");
        close(sockRaw);
        return -1L;
    }
    str=(*env)->GetStringUTFChars(env,addrString,NULL);
    if(str==NULL){
        printf("Error: Invalid addrString!\n");
        close(sockRaw);
        return -1L;
    }
    sockRaw=socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);
    if(sockRaw<0){
        printf("Error: Cannot create Socket\n");
        return -1L;
    }
    SET_NONBLOCKING(sockRaw);
//printf("SEND: %s WITH ID %d\n",str,Id);
    timeout=timeoutMillis+GetTickCount();
    bytesSent=sendPacket(str,Id,sockRaw);
//printf("SENT: %d BYTES\n",bytesSent);
    if(bytesSent<0){
        printf("Error: Cannot send packet\n");
        close(sockRaw);
        (*env)->ReleaseStringUTFChars(env, addrString,str );
        return -1L;
    }
    if(bytesSent!=(int)PACKET_SIZE+sizeof(ICMP_HEADER)){
        printf("Error: Too less bytes sent\n");
        close(sockRaw);
        (*env)->ReleaseStringUTFChars(env, addrString,str );
        return -1L;
    }
    rc=-2L;
    tv.tv_sec=timeoutMillis/1000;
    tv.tv_usec=(timeoutMillis%1000)*1000;
    while(rc==-2L){    //while receiving foreign packets
        FD_ZERO(&rd);
        FD_ZERO(&wr);
        FD_SET(sockRaw,&rd);
        if(select(sockRaw+1,&rd,NULL,NULL,&tv)>0){
            bytesReceived=recvfrom(sockRaw,recvbuf,sizeof(recvbuf),0,(struct sockaddr *)&addrFrom,&addrFromLen);
            remainingTime=timeout-GetTickCount();
//printf("RECEIVED: %d BYTES\n",bytesReceived);
            tv.tv_sec=(remainingTime)/1000;
            tv.tv_usec=(remainingTime%1000)*1000;
            if(remainingTime<=0){
                close(sockRaw);
                (*env)->ReleaseStringUTFChars(env, addrString,str );
//printf("TIMEOUT\n");
                return -1L;
            }
            if(bytesReceived<0){
//printf("NEGATIV\n");
                rc=-2L;
                continue;
            }
            rc=DecodePacket(recvbuf,bytesReceived,(unsigned short) Id);
//if(rc==-2)printf("%s FOREIGN ID: %d\n",str,Id);
        }
        else{
            close(sockRaw);
            (*env)->ReleaseStringUTFChars(env, addrString,str );
//printf("NO SELECT\n");
            return -1L;
        }
    }
    close(sockRaw);
    (*env)->ReleaseStringUTFChars(env, addrString,str );
//printf("DELAY: %d ms\n",rc);
    return rc;
}
#ifdef __cplusplus
}
#endif
